@@ -25,6 +25,7 @@
 #include "miscobj.h"
 #include "khwin.h"
 #include "dirtree.h"
+#include "dirutil.h"
 #include "cstrlib.h"
 
 //#define DUMP_HTML	
@@ -34,6 +35,8 @@
 # include <sys/types.h>
 #endif
 
+#include CXX__HEADER_algorithm
+
 string	saveCwd;
 
 extern string	dirFile;
@@ -368,7 +371,7 @@ void	HyperTreeDraw::Draw(HyperDocument &html, WINDOW *pad)
 }
 
 /*************************************************************************
-	Find Match
+	Match Objects
 *************************************************************************/
 
 struct MatchInfo {
@@ -389,6 +392,22 @@ struct MatchInfoIsCurDir {
 	}
 };
 
+struct MatchInfoIsBookmarked {
+	bool operator()(MatchInfo &m) const {
+		return m.dir->bookmarked;
+	}
+};
+
+struct MatchInfoIsNotBookmarkedNorCurDir {
+	bool operator()(MatchInfo &m) const {
+		return !m.dir->bookmarked && !m.is_curdir;
+	}
+};
+
+/*************************************************************************
+	Find Match
+*************************************************************************/
+
 void	RecursiveFindMatchDir(list<MatchInfo> &match_list, sptr<DirectoryEntry> &dir, 
 			      const string &str, const string &current_dir,
 			      bool add_curdir)
@@ -438,6 +457,47 @@ size_t	FindMatchDir(list<MatchInfo> &match_list, const string &linktext,
 	return match_list.size();
 }
 
+/*************************************************************************
+	Find Bookmark
+*************************************************************************/
+
+void	RecursiveFindBookmarkedDir(list<MatchInfo> &match_list, sptr<DirectoryEntry> &dir, 
+				    const string &current_dir)
+{
+	if (dir->IsDir() && dir->isUnreadable != 1 &&
+	    dir->bookmarked) {			// Found
+
+		match_list.push_back(MatchInfo(current_dir, dir, false));
+	}
+
+	for (sptr_list<DirectoryEntry>::iterator iter = dir->subDir.begin();
+	     iter != dir->subDir.end(); ++iter) {
+
+			    			// For dirs
+		if ((*iter)->IsDir()) {
+			string new_dir = current_dir;
+			if (new_dir[new_dir.size()-1] != '/')
+				new_dir += '/';
+			new_dir += (*iter)->GetNameStr();
+			
+			RecursiveFindBookmarkedDir(match_list, *iter, new_dir);
+		}
+	}
+}
+
+size_t	FindBookmarkedDir(list<MatchInfo> &match_list)
+{
+	match_list.clear();	
+	
+	for (sptr_list<DirectoryEntry>::iterator iter = dirTree.begin();
+	     iter != dirTree.end(); ++iter) {
+		string new_dir = (*iter)->GetNameStr();
+		RecursiveFindBookmarkedDir(match_list, *iter, new_dir);
+	}
+	
+	return match_list.size();
+}
+
 /*************************************************************************
 	Generate HTML-like codes from the match list
 *************************************************************************/
@@ -800,7 +860,7 @@ void	DirListCD(list<MatchInfo> &match_list, const string &init_pos,
 	fclose(file);
 }
 
-void	CD(const string &linktext)
+void	CD(const string &linktext, bool entire_tree)
 {
 	LoadDirFile(1);
 
@@ -823,18 +883,29 @@ void	CD(const string &linktext)
 							// to aid cycling through
 							// all match directory
 							// in the command line
-		size_t	ret = FindMatchDir(match_list, linktext, true);
+		FindMatchDir(match_list, linktext, true);
 
-		size_t	count_adj = 0;
+		if (!entire_tree
+		    && count_if(match_list.begin(), match_list.end(),
+			        MatchInfoIsBookmarked())) {
+							// Remove entries not
+							// bookmarked
+			match_list.remove_if(MatchInfoIsNotBookmarkedNorCurDir());
+		}
+
+		int	adjust = 0;			// In case getcwd fails
 		for (list<MatchInfo>::iterator iter = match_list.begin();
 		     iter != match_list.end(); ++iter) {
 			if ((*iter).is_curdir) {
-				count_adj = 1;
+				adjust = 1;
 				break;
 			}
 		}
 
-		if (ret > static_cast<size_t>(kcdConfig.cfgShowListThreshold)+count_adj) {
+							// +1 because of current
+							// directory in match_list
+		if (match_list.size() > static_cast<size_t>(kcdConfig.cfgShowListThreshold)
+					+adjust) {
 
 							// Show whole list
 
@@ -929,3 +1000,212 @@ void	CD(const string &linktext)
 	}
 }
 
+void	ShowBookmark(const string &linktext)
+{
+	LoadDirFile(1);
+
+	Buffer	buffer;
+
+	HyperDraw *drawObj = 0;
+
+	string	hyperInitPos;
+	string	hyperFindLink;
+	
+	GenerateHTMLStrings();
+
+	list<MatchInfo>	match_list;
+	if (linktext.size()) {
+		FindMatchDir(match_list, linktext);
+							// Remove entries not
+							// bookmarked
+		match_list.remove_if(MatchInfoIsNotBookmarkedNorCurDir());
+	}
+	else {
+		FindBookmarkedDir(match_list);
+	}
+
+	if (!match_list.size()) {
+		cout << flush;			// Avoid out-of-order display
+		gtout(cerr, _("%$: no bookmarks found\n"))
+		     << progName;
+		return;
+	}
+
+	drawObj = new HyperListDraw(match_list);
+			
+	hyperInitPos = "";
+	hyperFindLink = "";
+			
+	HyperDocument		*html = new HyperDocument(drawObj);
+
+	{
+						// Display
+		NCurses::Init();
+
+		try {
+
+			if (COLS < 50 || LINES < 10) {
+				throw ErrorScreenTooSmall();
+			}
+			else {
+				init_entity_table();
+
+				khGeometryManager	geoMan(modeHVScroll);
+				khScreenManager		scrnMan(geoMan);
+
+				khStatusWindowWithFind	findBar(scrnMan, idStatWindow, NULL);
+				khURLWindowWithFind	urlBar(scrnMan, idURLWindow);
+				khHScrollBar		hscroll(scrnMan, idHScrollBar);
+				khVScrollBar		vscroll(scrnMan, idVScrollBar);
+				khScrollBox		cursorRest(scrnMan, idScrollBox);
+	
+				urlBar.SetTitle(" ");		// Empty title
+				scrnMan.SetCursor(&cursorRest);
+				scrnMan.RequestUpdateAndRestCursor();
+
+							// Use pre-computed 
+							// document size
+				khHyperWindowWithFind	hyper(scrnMan, idHyperWindow,
+							      *html, findBar, urlBar,
+							      hscroll, vscroll, hyperInitPos,
+							      1, htmlRow, htmlCol);
+
+				findBar.SetFindWindow(&hyper);
+				hscroll.SetWin(&hyper);	// Handle scroll through mouse
+				vscroll.SetWin(&hyper);
+				
+							// Set cfgSpaceSelect flag
+				khStatusWindowWithFind::spaceSelect = kcdConfig.cfgSpaceSelect;
+
+							// Send all keyboard input
+							// to hyper
+				scrnMan.SetFocus(&hyper);
+			
+							// Paint screen
+				scrnMan.RequestUpdateAndRestCursor();
+				hyper.KeyboardLoop();
+			}
+		}
+		catch (...) {
+			move(LINES-1, 0);
+			printf("\n\n");		// Avoid printing error
+						// message on top of drawn
+						// screen
+			NCurses::End();
+			throw;
+		}
+		NCurses::End();
+	}
+}
+
+void	AddBookmark(const string &str)
+{
+	string s = ExpandDirectory(str, false);
+	FindDirInfo *f = FindDir(s, true, false);
+	if (f)
+		f->dir_ptr->bookmarked = true;
+	else {
+		cout << flush;			// Avoid out-of-order display
+		gtout(cerr, _("%$: cannot find %$ in saved directory tree\n"
+			      "        Directory ignored"))
+		     << progName << s;
+	}
+}
+
+void	RemoveBookmark(const string &str)
+{
+	string s = ExpandDirectory(str, false);
+	FindDirInfo *f = FindDir(s, true, false);
+	if (f)
+		f->dir_ptr->bookmarked = false;
+	else {
+		cout << flush;			// Avoid out-of-order display
+		gtout(cerr, _("%$: cannot find %$ in saved directory tree\n"
+			      "        Directory ignored"))
+		     << progName << s;
+	}
+}
+
+void	AddBookmarks(int argc, char *argv[], int argcBegin)
+{
+	LoadDirFile(1);
+
+	if (argcBegin == argc) {
+		if (saveCwd.size())
+			AddBookmark(saveCwd);
+		else {
+			cout << flush;			// Avoid out-of-order display
+			gtout(cerr, _("%$: cannot determine current directory\n"
+				      "        Directory ignored"))
+			     << progName;
+			return;
+		}
+	}
+	else {
+		for (int i = argcBegin; i < argc; ++i) {
+			if (!argv[i][0])
+				continue;
+
+			if (argv[i][0] != '/' && argv[i][0] != '~') {
+				if (saveCwd.size()) {
+					string dir = saveCwd;
+					dir += '/';
+					dir += argv[i];
+					AddBookmark(dir);
+				}
+				else {
+					cout << flush;			// Avoid out-of-order display
+					gtout(cerr, _("%$: cannot determine current directory for path %$\n"
+						      "        Directory ignored"))
+					     << progName << argv[i];
+				}
+			}
+			else
+				AddBookmark(argv[i]);
+		}
+	}
+
+	WriteDirFile();
+}
+
+void	RemoveBookmarks(int argc, char *argv[], int argcBegin)
+{
+	LoadDirFile(1);
+
+	if (argcBegin == argc) {
+		if (saveCwd.size())
+			RemoveBookmark(saveCwd);
+		else {
+			cout << flush;			// Avoid out-of-order display
+			gtout(cerr, _("%$: cannot determine current directory\n"
+				      "        Directory ignored"))
+			     << progName;
+			return;
+		}
+	}
+	else {
+		for (int i = argcBegin; i < argc; ++i) {
+			if (!argv[i][0])
+				continue;
+
+			if (argv[i][0] != '/' && argv[i][0] != '~') {
+				if (saveCwd.size()) {
+					string dir = saveCwd;
+					dir += '/';
+					dir += argv[i];
+					RemoveBookmark(dir);
+				}
+				else {
+					cout << flush;			// Avoid out-of-order display
+					gtout(cerr, _("%$: cannot determine current directory for path %$\n"
+						      "        Directory ignored"))
+					     << progName << argv[i];
+				}
+			}
+			else
+				RemoveBookmark(argv[i]);
+		}
+	}
+
+	WriteDirFile();
+}
