@@ -41,6 +41,10 @@ struct FindDirInfo {
 FindDirInfo *FindDir(const string &str, bool honor_cfg, bool change = true);
 FindDirInfo *FindDir(const string &str, bool honor_cfg, bool change_top, 
 		     bool change_sub);
-void	CD(const string &linktext);
+void	CD(const string &linktext, bool entire_tree = false);
+void	ShowBookmark(const string &linktext);
+
+void	AddBookmarks(int argc, char *argv[], int argcBegin);
+void	RemoveBookmarks(int argc, char *argv[], int argcBegin);
 
 #endif	/* __K_GENTREE_H */
