@@ -91,9 +91,14 @@ void	PrintBanner()
 void	Help()
 {
 	PrintBanner();
-	gtout(cout, _("Usage:   %$ [DIR]\n"
+	gtout(cout, _("Usage:   %$ [PATTERN]\n"
 		      "  or:    %$ [OPTION]\n")) << progName << progName;
 	cout << _("Available options:\n");
+	cout << _("     -b [PATTERN]   Display bookmarked directories matching PATTERN\n"
+		  "                    (default = display all)\n"
+		  "     -ba [DIR ...]  Bookmark DIR list (default = current directory)\n"
+		  "     -br [DIR ...]  Remove DIR list bookmark\n"
+		  "     -B             Use entire tree even a bookmarked match is found\n");
 	cout << _("     --config       Display current configuration\n"
 		  "     --configattr   Display current screen attribute configuration\n"
 		  "     --features     Display features included during compilation\n"
@@ -261,13 +266,15 @@ void	mainReal(int argc, char *argv[])
 	if (argc == 1) { 		// No parameters
 		SetSaveFile();
 
-		CD("");
+		CD("");			// Show entire tree
 	}
 	else {
 		int	argcCur = 1;
 		int	argcLeft = argc-1;
 
-		int	rescan = 0;
+		bool	rescan = false;
+		bool	entire_tree = false;
+
 		vector<string>	partial_dir;
 
 		while (argcLeft) {
@@ -345,14 +352,14 @@ void	mainReal(int argc, char *argv[])
 				return;
 			}
 			else if (strcmp(argv[argcCur], "-iau") == 0) {
-				throw ErrorGenericCommandLine(_("user-specific installation depreciated"));
+				throw ErrorGenericCommandLine(_("user-specific installation deprecated"));
 			}
 			else if (strcmp(argv[argcCur], "-icu") == 0) {
-				throw ErrorGenericCommandLine(_("user-specific installation depreciated"));
+				throw ErrorGenericCommandLine(_("user-specific installation deprecated"));
 				return;
 			}
 			else if (strcmp(argv[argcCur], "-ifu") == 0) {
-				throw ErrorGenericCommandLine(_("user-specific installation depreciated"));
+				throw ErrorGenericCommandLine(_("user-specific installation deprecated"));
 				return;
 			}
 			else if (strcmp(argv[argcCur], "-ias") == 0) {
@@ -400,10 +407,38 @@ void	mainReal(int argc, char *argv[])
 						// Consume argument of -rp
 				argcCur++;
 				argcLeft--;
+			}
+						// Show bookmark
+			else if (strcmp(argv[argcCur], "-b") == 0) {
+				SetSaveFile();
+
+				if (argcLeft >= 3)
+					throw ErrorGenericCommandLine(_("too many arguments for `-b\'"));
+				else if (argcLeft == 2)
+					ShowBookmark(argv[argcCur+1]);
+				else
+					ShowBookmark("");
+				return;
+			}
+						// Add bookmark
+			else if (strcmp(argv[argcCur], "-ba") == 0) {
+				SetSaveFile();
+				AddBookmarks(argc, argv, argcCur+1);
+				return;
+			}
+						// Remove bookmark
+			else if (strcmp(argv[argcCur], "-br") == 0) {
+				SetSaveFile();
+				RemoveBookmarks(argc, argv, argcCur+1);
+				return;
+			}
+						// Use entire tree
+			else if (strcmp(argv[argcCur], "-B") == 0) {
+				entire_tree = true;
 			}
 						// Rescan dir.
 			else if (strncmp(argv[argcCur], "-r", 2) == 0) {
-				rescan = 1;
+				rescan = true;
 
 				bool	use_partial_dir = false;
 
@@ -449,8 +484,8 @@ void	mainReal(int argc, char *argv[])
 			else {				// Not an option, must be a directory name
 				SetSaveFile();
 
-				if (argcLeft == 1 && rescan == 0)
-					CD(argv[argcCur]);
+				if (argcLeft == 1 && rescan == false)
+					CD(argv[argcCur], entire_tree);
 				else			// Too many arguments
 					throw ErrorGenericCommandLine(_("too many arguments"));
 				return;
@@ -472,7 +507,7 @@ void	mainReal(int argc, char *argv[])
 							// no action specified yet
 			SetSaveFile();
 
-			CD("");
+			CD("", entire_tree);		// Show entire tree
 		}
 	}
 }
