@@ -27,7 +27,7 @@
 #include "editbox.h"
 #include "khdoc.h"		/* For attributes */
 
-USING_NAMESPACE_STD
+USING_NAMESPACE_STD;
 
 /*************************************************************************
 	Status window base class
