@@ -327,15 +327,18 @@ size_t	GetWStringWidth(const k_wstring &s, size_t n)
 		return (n > ssize) ? ssize : n;
 }
 
-size_t	GetWStringWidthFirstCharAtPos(const k_wstring &s, size_t pos)
+size_t	GetWStringLastCharAtPos(const k_wstring &s, size_t pos)
 {
 	size_t ssize = s.size();
 #ifdef USE_UTF8_MODE
 	if (IsUTF8Mode()) {
 		size_t i = 0;
 		for (size_t size = 0; i < ssize; ++i) {
-			if (size >= pos)
+			if (size >= pos) {
+				while (i < ssize && wcwidth(s[i]) == 0)
+					++i;
 				return i;
+			}
 			size += wcwidth(s[i]);
 		}
 		return i;
@@ -345,3 +348,7 @@ size_t	GetWStringWidthFirstCharAtPos(const k_wstring &s, size_t pos)
 		return (pos > ssize) ? ssize : pos;
 }
 
+size_t	GetWStringWidthLastCharAtPos(const k_wstring &s, size_t pos)
+{
+	return GetWStringWidth(s, GetWStringLastCharAtPos(s, pos));
+}
